-------------------------------------------------------------------------------
-- Animals Mod by Sapier
-- 
-- You may copy, use, modify or do nearly anything except removing this
-- copyright notice. 
-- And of course you are NOT allow to pretend you have written it.
--
-- (c) Sapier
-- Contact sapier a t gmx net
-------------------------------------------------------------------------------

--registry for movement patterns
animals_movement_patterns = {}
animals_defined = {}
animals_spawn_algorithms = {}

animals_modpath = minetest.get_modpath("animals")

--include debug trace functions
dofile (animals_modpath .. "/debug_trace.lua")

--include engine
dofile (animals_modpath .. "/generic_functions.lua")
dofile (animals_modpath .. "/movement_gen.lua")
dofile (animals_modpath .. "/harvesting.lua")
dofile (animals_modpath .. "/weapons.lua")
dofile (animals_modpath .. "/fighting.lua")
dofile (animals_modpath .. "/random_drop.lua")
dofile (animals_modpath .. "/permanent_data.lua")
dofile (animals_modpath .. "/management_functions.lua")
dofile (animals_modpath .. "/debug.lua")

--include spawning support
dofile (animals_modpath .. "/spawning.lua")

local version = "1.0.0"

LOGLEVEL_INFO     = "verbose"
LOGLEVEL_NOTICE   = "info"
LOGLEVEL_WARNING  = "action"
LOGLEVEL_ERROR    = "error"
LOGLEVEL_CRITICAL = "error"


--define tools used for more than one animal
minetest.register_craft({
	output = "animalmaterials:lasso 5",
	recipe = {
		{'', "animalmaterials:wool_white",''},
		{"animalmaterials:wool_white",'', "animalmaterials:wool_white"},
		{'',"animalmaterials:wool_white",''},
	}
})

minetest.register_craft({
	output = "animalmaterials:net 1",
	recipe = {
		{"animalmaterials:wool_white",'',"animalmaterials:wool_white"},
		{'', "animalmaterials:wool_white",''},
		{"animalmaterials:wool_white",'',"animalmaterials:wool_white"},
	}
})


function animals_init_module()
	print("Initializing animals mod")

	print("Initializing movement patterns..")
	animals_register_movement_patterns()

	print("Initializing weaponry..")
	animals_init_weapons()

	print("Initializing debug hooks..")
	animals_init_debug()

	print("animals mod "..version.." loaded")
end

animals_init_module()


--temporary cleanup code for old animals

minetest.register_entity(":animals:sheep",
	{	
		on_step = function(self,dtime)
			self.object:remove()
			end,
		on_activate = function(self,staticdata)
			self.object:remove()
			end
	})
	
minetest.register_entity(":animals:sheep_naked",
	{
		on_step = function(self,dtime)
			self.object:remove()
			end,
		on_activate = function(self,staticdata)
			self.object:remove()
			end
	})
	
minetest.register_entity(":animals:cow",
	{
		on_step = function(self,dtime)
			self.object:remove()
			end,
		on_activate = function(self,staticdata)
			self.object:remove()
			end
	})
	
minetest.register_entity(":animals:deer",
	{
		on_step = function(self,dtime)
			self.object:remove()
			end,
		on_activate = function(self,staticdata)
			self.object:remove()
			end
	})
	
minetest.register_entity(":animals:dm",
	{
		on_step = function(self,dtime)
			self.object:remove()
			end,
		on_activate = function(self,staticdata)
			self.object:remove()
			end
	})
	
minetest.register_entity(":animals:big_red",
	{
		on_step = function(self,dtime)
			self.object:remove()
			end,
		on_activate = function(self,staticdata)
			self.object:remove()
			end
	})
	
minetest.register_entity(":animals:creeper",
	{
		on_step = function(self,dtime)
			self.object:remove()
			end,
		on_activate = function(self,staticdata)
			self.object:remove()
			end
	})
	
minetest.register_entity(":animals:chicken",
	{
		on_step = function(self,dtime)
			self.object:remove()
			end,
		on_activate = function(self,staticdata)
			self.object:remove()
			end
	})
	
minetest.register_entity(":animals:fish_blue_white",
	{
		on_step = function(self,dtime)
			self.object:remove()
			end,
		on_activate = function(self,staticdata)
			self.object:remove()
			end
	})
	
minetest.register_entity(":animals:rat",
	{
		on_step = function(self,dtime)
			self.object:remove()
			end,
		on_activate = function(self,staticdata)
			self.object:remove()
			end
	})
	
minetest.register_entity(":animals:vampire",
	{
		on_step = function(self,dtime)
			self.object:remove()
			end,
		on_activate = function(self,staticdata)
			self.object:remove()
			end
	})

minetest.register_abm({
		nodenames = { "animals:sheep_wool" },
		interval = 1,
		chance = 1,

		action = function(pos, node, active_object_count, active_object_count_wider)
			minetest.env:remove_node(pos)
			minetest.env:add_node(pos,{name="animalmaterials:wool_white"})
		end
	
	})

