-------------------------------------------------------------------------------
-- Animals Mod by Sapier
-- 
-- You may copy, use, modify or do nearly anything except removing this
-- copyright notice. 
-- And of course you are NOT allow to pretend you have written it.
--
-- (c) Sapier
-- Contact sapier a t gmx net
-------------------------------------------------------------------------------

movement_gen = {}


--movement generator prediction time
prediction_time = 2

mov_patterns_defined = {}

dofile (animals_modpath .. "/movement_patterns/dont_move.lua")
dofile (animals_modpath .. "/movement_patterns/run_and_jump_low.lua")
dofile (animals_modpath .. "/movement_patterns/stop_and_go.lua")
dofile (animals_modpath .. "/movement_patterns/swim_pattern1.lua")
dofile (animals_modpath .. "/movement_patterns/swim_pattern2.lua")
dofile (animals_modpath .. "/movement_patterns/flight_pattern1.lua")



function animals_register_movement_patterns() 
		for index,value in ipairs(mov_patterns_defined) do 

		print ("\tregistering pattern "..index.." -> "..value.name)
		animals_movement_patterns[value.name] = value
	end
end

-------------------------------------------------------------------------------
-- name: is_jumpable_surface(name)
--
-- action: check if name is a surface an animal may jump onto
--
-- param1: name to check
-- retval: true/false
-------------------------------------------------------------------------------
function movement_gen.is_jumpable_surface(name)

	
	if 	name == "default:dirt" or
		name == "default:dirt_with_grass" or
		name == "default:stone" or
		name == "default:sand" or
		name == "default:clay"
		then
		return true
	end

	dbg_animals.movement_lvl1("ANIMALS: is "..name.." a jumpable surface?")
	return false
end

-------------------------------------------------------------------------------
-- name: pos_is_ok(pos,entity)
--
-- action: check if a position is suitable for an animal
--
-- param1: position to check
-- retval: suitability of position for animal
--           -ok                    -->position is ok
--           -collision             -->position is within a node
--           -collision_jumpable    -->position is within a node that can be jumped onto
--           -drop                  -->position is a drop
--           -drop_above_water      -->position is to far above water
--           -above_water           -->position is right over water
--           -in_water              -->position is within a water node(source or flow)
--			 -in_air                -->position is in air
--           -invalid               -->unable to check position
-------------------------------------------------------------------------------
function movement_gen.pos_is_ok(pos,entity)

	local min_ground_distance   = 0
	local max_ground_distance   = 0

	if entity.animals_mpattern.movement_canfly == false then
		max_ground_distance = 1	
	end
	


	dbg_animals.movement_lvl1("Checking pos "..printpos(pos))

	if pos == nil then
		minetest.log(LOGLEVEL_ERROR,"ANIMALS: BUG!!!! checking pos with nil value this won't work")
		return "invalid"	
	end
	
	local node = minetest.env:get_node(pos)

	if node == nil then
		minetest.log(LOGLEVEL_ERROR,"ANIMALS: BUG!!!! checking position with invalid node")
		return "invalid"
	end

	if animals_is_in_environment(node.name,entity.animals_mpattern.movement_environment) then

			--following return codes are only usefull for non flying
			if entity.animals_mpattern.movement_canfly == false then

				if animals_above_water(pos) then
					if ground_distance > regular_ground_distance then
						return "drop_above_water"
					end

					return "above_water"
				end

				local ground_distance = animals_ground_distance(pos)

				if ground_distance > max_ground_distance then
					return "drop"
				end
			end		
	
			--TODO add check if in correct heigth level?
			return "ok"
	end

	--position is not ok gather some usefull information
	local pos_above = {x=pos.x,y=pos.y+1,z=pos.z}	
	local node_above = minetest.env:get_node(pos_above)

	if node_above == nil then
		minetest.log(LOGLEVEL_ERROR,"ANIMALS: BUG!!!! checking position with invalid node above")
		return "invalid"
	end

	if 	animals_is_in_environment(node_above.name,entity.animals_mpattern.movement_environment) and
		movement_gen.is_jumpable_surface(node.name) then
		return "collision_jumpable"
	end

	if node.name == "default:water_source" or 
		node.name == "default:water_flowing" then
		return "in_water"
	end

	if node.name == "air" then
		return "in_air"
	end

	return "collision"

end

-------------------------------------------------------------------------------
-- name: movement_gen.get_suitable_pos_same_level(pos_raw,maxsearcharea,entity)
--
-- action: find a position suitable around a specific position
--
-- param1: position to look at
-- param2: max range to look for suitable position
-- retval: position found
-------------------------------------------------------------------------------
function movement_gen.get_suitable_pos_same_level(pos_raw,maxsearcharea,entity)

	local pos = animals_round_pos(pos_raw)

	dbg_animals.movement_lvl1("ANIMALS: Starting pos is "..printpos(pos).." max search area is "..maxsearcharea)

	local e1 = "|"
	local e2 = "|"
	local e3 = "|"
	local e4 = "|"

	--search next position on solid ground
	for search=1, maxsearcharea,1 do
		--TODO randomize search order

		--find along edge 1
		for current=-maxsearcharea,maxsearcharea,1 do
			local pos_tocheck = { x= pos.x + current,y=pos.y,z=pos.z -maxsearcharea}		
			local pos_state = movement_gen.pos_is_ok(pos_tocheck,entity)

			dbg_animals.movement_lvl1("ANIMALS: state of "..printpos(pos_tocheck).." is ".. pos_state)

			if pos_state == "ok" then
				dbg_animals.movement_lvl1("found new pos")
				return pos_tocheck
			else
				e1 = e1..pos_state.."|"
			end			
		end

		--find along edge 2
		for current=-maxsearcharea,maxsearcharea,1 do
			local pos_tocheck = { x= pos.x + maxsearcharea,y=pos.y,z=pos.z + current}		
			local pos_state = movement_gen.pos_is_ok(pos_tocheck,entity)
			
			dbg_animals.movement_lvl1("ANIMALS: state of "..printpos(pos_tocheck).." is ".. pos_state)

			if pos_state == "ok" then
				dbg_animals.movement_lvl1("found new pos")
				return pos_tocheck
			else
				e2 = e2..pos_state.."|"
			end			
		end

		--find along edge 3

		for current=-maxsearcharea,maxsearcharea,1 do
			local pos_tocheck = { x= pos.x + current,y=pos.y,z=pos.z + maxsearcharea}		
			local pos_state = movement_gen.pos_is_ok(pos_tocheck,entity)

			dbg_animals.movement_lvl1("ANIMALS: state of "..printpos(pos_tocheck).." is ".. pos_state)

			if pos_state == "ok" then
				dbg_animals.movement_lvl1("found new pos")
				return pos_tocheck
			else
				e3 = e3..pos_state.."|"
			end			
		end

		--find along edge 4
		for current=-maxsearcharea,maxsearcharea,1 do
			local pos_tocheck = { x= pos.x -maxsearcharea,y=pos.y,z=pos.z + current}		
			local pos_state = movement_gen.pos_is_ok(pos,entity)

			dbg_animals.movement_lvl1("ANIMALS: state of "..printpos(pos_tocheck).." is ".. pos_state)

			if pos_state == "ok" then
				dbg_animals.movement_lvl1("found new pos")
				return pos_tocheck
			else
				e4 = e4..pos_state.."|"
			end			
		end
	end

--	print("ANIMALS: Bug !!! didn't find a suitable position to place animal")
--	print("Surrounding was:")
--	print(e1)
--	print(e2)
--	print(e3)
--	print(e4)
	return nil
end


-------------------------------------------------------------------------------
-- name: get_random_acceleration(minaccel,maxaccel)
--
-- action: get a random x/z acceleration within a specified acceleration range
--
-- param1: minimum acceleration to use
-- param2: maximum acceleration
-- retval: x/y/z acceleration
-------------------------------------------------------------------------------
function movement_gen.get_random_acceleration(minaccel,maxaccel)

	--calc random absolute value
	local rand_x = (math.random() * (maxaccel - minaccel)) + minaccel
	local rand_z = (math.random() * (maxaccel - minaccel)) + minaccel

	--randomize direction

	if math.random() < 0.5 then
		rand_x = rand_x * -1
	end

	if math.random() < 0.5 then
		rand_z = rand_z * -1
	end

	return {
		x = rand_x,
		y = nil,
		z = rand_z
		}
end

-------------------------------------------------------------------------------
-- name: movement_gen.calc_new_pos(pos,acceleration,prediction_time)
--
-- action: calc the position an animal would be after a specified time
--         this doesn't handle velocity changes due to colisions
--
-- param1: current position
-- param2: acceleration to be set
-- param3: time to pass
-- param4: current velocity
-- retval: position after specified time
-------------------------------------------------------------------------------
function movement_gen.calc_new_pos(pos,acceleration,prediction_time,current_velocity)	

	local predicted_pos = {x=pos.x,y=pos.y,z=pos.z}

	predicted_pos.x = predicted_pos.x + current_velocity.x * prediction_time + (acceleration.x/2)*math.exp(prediction_time,2)
	predicted_pos.z = predicted_pos.z + current_velocity.z * prediction_time + (acceleration.z/2)*math.exp(prediction_time,2)


	return predicted_pos
end



-------------------------------------------------------------------------------
-- name: movement_gen.get_accel_to(old_pos,entity) 
--
-- action: calculate a random speed directed from old_pos to new_pos
--
-- param1: source position
-- param2: entity
-- retval: x/y/z speed
-------------------------------------------------------------------------------
--
function movement_gen.get_accel_to(new_pos,entity)

	if new_pos == nil or entity == nil then
		minetest.log(LOGLEVEL_CRITICAL,"ANIMALS: movement_gen.get_accel_to : Invalid parameters")
	end
	
	local old_pos  = entity.object:getpos()
	local node 	   = minetest.env:get_node(old_pos)
	local maxaccel = entity.data.movement.max_accel
	local minaccel = entity.data.movement.min_accel
	
	local yaccel = animals_get_default_acceleration(old_pos,
							entity.animals_mpattern.movement_environment,
							entity.animals_mpattern.movement_canfly)

	--TODO calc y speed for flying animals

	local x_diff = new_pos.x - old_pos.x
	local z_diff = new_pos.z - old_pos.z

	local rand_x = (math.random() * (maxaccel - minaccel)) + minaccel
	local rand_z = (math.random() * (maxaccel - minaccel)) + minaccel

	if x_diff < 0 then
		rand_x = rand_x * -1
	end

	if z_diff < 0 then
		rand_z = rand_z * -1
	end

	return { x=rand_x,y=yaccel,z=rand_z }
end

-------------------------------------------------------------------------------
-- name: movement_gen.changeaccel(pos,prediction_time,entity)
--
-- action: find a suitable new acceleration for animal
--
-- param1: current position
-- param2: prediction time
-- param3: entity to change accel for
-- retval: x/y/z accel + jump flag
-------------------------------------------------------------------------------
function movement_gen.changeaccel(pos,prediction_time,entity)

	local current_velocity = entity.object:getvelocity()

	local new_accel = movement_gen.get_random_acceleration(entity.data.movement.min_accel,
														entity.data.movement.max_accel)

	local pos_predicted = movement_gen.calc_new_pos(pos,new_accel,prediction_time,current_velocity)

	local maxtries = 5

	local state = movement_gen.pos_is_ok(pos_predicted,entity)

	while  state ~= "ok" do
		local done = false

		--don't loop forever get to save mode and try next time
		if maxtries < 0 then
			dbg_animals.movement_lvl1("ANIMALS: Aborting acceleration finding for this cycle due to max retries")
			if state == "collision" or
				state == "collision_jumpable" then
				dbg_animals.movement_lvl1("Returning "..printpos(new_accel).." as new accel")
				return new_accel
			end
		
			dbg_animals.movement_lvl1("ANIMALS: Didn't find a suitable acceleration: " .. entity.data.name .. printpos(pos))
			return nil
		end

		--first try to invert acceleration on collision
		if not done and (state == "collision" or
			state == "collision_jumpable") then
			new_accel = { 	x= new_accel.x * -1,
					y= new_accel.y,
					z= new_accel.z * -1}

			pos_predicted = movement_gen.calc_new_pos(pos,new_accel,prediction_time,current_velocity)
			if movement_gen.pos_is_ok(pos_predicted,entity) == "ok" then
				done = true
			end

		end

		--generic way to find new acceleration
		if not done then
			new_accel = movement_gen.get_random_acceleration(entity.data.movement.min_accel,
														entity.data.movement.max_accel)
		
			pos_predicted = movement_gen.calc_new_pos(pos,new_accel,prediction_time,current_velocity)
		end

		state = movement_gen.pos_is_ok(pos_predicted,entity)
		maxtries = maxtries -1
	end

	return new_accel

end


-------------------------------------------------------------------------------
-- name: movement_gen.fix_current_pos(entity)
--
-- action: check current position for validity and try to fix
--
-- param1: entity to generate movement
-- retval: {
--		speed_to_set = {x,y,z}
--		speed_changed = true/false
--              force_speed_change = true/false
--	}
-------------------------------------------------------------------------------
function movement_gen.fix_current_pos(entity,pos)

	local pos = entity.getbasepos(entity)
	
	--check if current pos is ok
	local current_state = movement_gen.pos_is_ok(pos,entity)

	local default_y_accel = animals_get_default_acceleration(pos,
								entity.animals_mpattern.movement_environment,
								entity.animals_mpattern.movement_canfly)

	local accel_changed = false

	local accel_to_set = { x=0,
							y=default_y_accel,
							z=0 }

	if entity.animals_acceleration ~= nil then
		accel_to_set = entity.animals_acceleration
	end

	local force_change_of_accel = false
	local abort_processing = false

	if current_state ~= "ok" and
		current_state ~= "drop" then
		dbg_animals.movement_lvl1("ANIMALS: BUG !!! somehow your animal managed to get where it shouldn't be, trying to fix")

		--stop animal from moving at all
		entity.object:setacceleration({x=0,y=default_y_accel,z=0})
		force_change_of_accel = true

		--animal is currently in whater try to find a suitable position 1 level above current level
		if current_state == "in_water" then
			local targetpos = movement_gen.get_suitable_pos_same_level({x=pos.x,y=pos.y+1,z=pos.z},1,entity)

			if targetpos ~= nil then
				minetest.log(LOGLEVEL_WARNING,"ANIMALS: Your animal dropt into water moving to "..
						printpos(targetpos).." state: "..
						movement_gen.pos_is_ok(targetpos,entity))
				
				entity.object:moveto(targetpos)
				accel_to_set.y = animals_get_default_acceleration(targetpos,
								entity.animals_mpattern.movement_environment,
								entity.animals_mpattern.movement_canfly)
			else
				minetest.log(LOGLEVEL_ERROR,"ANIMALS: BUG !!! didn't find a way out of water, drowning animal")
				abort_processing = true
		
				spawning.remove(entity)
			end
		end

		if current_state == "in_air" then
			--TODO die?
		end

		--animal is most likely to fall in water soon find a dry place around
		if current_state == "above_water" then
			dbg_animals.movement_lvl1("ANIMALS: animal is lurking around above water")
			local targetpos = movement_gen.get_suitable_pos_same_level({x=pos.x,y=pos.y,z=pos.z},1,entity)
	
			--try a large jump
			--if targetpos == nil then
			--	local targetpos = movement_gen.get_suitable_pos_same_level({x=pos.x,y=pos.y+1,z=pos.z},1,entity)
			--end

			if targetpos ~= nil then
--simple way of fixing
				dbg_animals.movement_lvl1("Animal ".. entity.animals_name .. " above water, moving to "..printpos(targetpos).." state: "..movement_gen.pos_is_ok(targetpos,entity))
				entity.object:moveto(targetpos)	

--more sophisticated way of fixing... but doesn't work right now		
--				entity.object:moveto(animals_fix_y_pos(pos))	
--				speed_to_set =  animals_get_speed_to(pos,targetpos,
--									entity.animals_min_speed,
--									entity.animals_max_speed)
--				speed_to_set.y = 0
--				speed_changed = true
			else
				minetest.log(LOGLEVEL_ERROR,"ANIMALS: BUG !!! didn't find a way from water, animal is most likely to drown soon")
			end
		end      
	end


	return {	accel = accel_to_set, 	
			accel_changed=accel_changed,
			force_accel_change=force_change_of_accel,
			abort_processing=abort_processing
			 }
end

-------------------------------------------------------------------------------
-- name: predict_next_block(pos,velocity,acceleration)
--
-- action: predict next block based on pos velocity and acceleration
--
-- param1: current position
-- param2: current velocity
-- param3: current acceleration
-- retval: position of next block animal will be
-------------------------------------------------------------------------------
function movement_gen.predict_next_block(pos,velocity,acceleration)

	local pos_predicted = movement_gen.calc_new_pos(pos,
								acceleration,
								prediction_time,
								velocity
								)
	local count = 1

	--check if after prediction time we would have traveled more than one block and adjust to not predict to far
	while animals_calc_distance(pos,pos_predicted) > 1 do		
		dbg_animals.movement_lvl1("Animal will move more than one block within prediction period ")
		
		pos_predicted = movement_gen.calc_new_pos(pos,
								acceleration,
								prediction_time - (count*0.1),
								velocity
								)

		if (prediction_time - (count*0.1)) < 0 then
			minetest.log(LOGLEVEL_WARNING,"ANIMALS: Bug!!!! didn't find a suitable prediction time. Animal will move more than one block within prediction period: ".. distance_to_predicted_pos)
			break
		end

		count = count +1
	end
	
	return pos_predicted

end

-------------------------------------------------------------------------------
-- name: handle_level_change_end(entity,default_accel,current_velocity,current_acceleration,now)
--
-- action: check if there is a level change in progress that may 
-- need to be stopped
--
-- param1: entity to generate movement
-- retval: -
-------------------------------------------------------------------------------
function movement_gen.handle_level_change_end(entity,default_accel,current_velocity,current_acceleration,now,accel_to_set)

	if entity.animals_mpattern.movement_canfly ~= nil and
		entity.dynamic_data.movement.changing_levels == true then
		
		local level_change_time = movement_gen.calc_level_change_time(entity,default_accel)

		local time_completed = entity.dynamic_data.movement.ts_random_jump + level_change_time

		dbg_animals.movement_lvl1("ANIMALS: ".. now .. " " .. entity.data.name .. 
			" check complete level change " .. time_completed)

		if entity.dynamic_data.movement.changing_levels and
			time_completed  < now then

			dbg_animals.movement_lvl2("ANIMALS: ".. now .. " " .. entity.data.name .. 
			" level change complete reestablishing default y acceleration " .. default_accel)
			entity.dynamic_data.movement.changing_levels = false
		
			current_velocity.y = 0
			entity.object:setvelocity(current_velocity)
		
			local accel_to_set = current_acceleration
			accel_to_set.y = default_accel
			return true, accel_to_set
		end
	end
	
	return false, accel_to_set
end

-------------------------------------------------------------------------------
-- name: calc_level_change_time(entity)
--
-- action: main movement generation function
--
-- param1: entity to generate movement
-- retval: -
-------------------------------------------------------------------------------
function movement_gen.calc_level_change_time(entity,default_accel) 
	local retval = 1 --default value

	--calculate a reasonable value to stop level change
	if entity.animals_mpattern.movement_canfly == false then --case animal can't fly 
		return 0
	else
		-- if it's a flying animal and left it's normal medium
		if default_accel ~= 0 then 
			retval = 0			
		else
			retval = math.sqrt(2/entity.data.movement.min_accel)
		end
	end

	return retval
end
-------------------------------------------------------------------------------
-- name: generator(entity)
--
-- action: main movement generation function
--
-- param1: entity to generate movement
-- retval: -
-------------------------------------------------------------------------------
function movement_gen.generator(entity)

	if entity == nil then
		return
	end

	--check current position	
	local retval = movement_gen.fix_current_pos(entity)

	--animal has been removed due to unfixable problems
	if retval.abort_processing then		
		return
	end

	local centerpos             = entity.object:getpos()
	local pos 					= entity.getbasepos(entity)
	
	local current_acceleration 	= entity.object:getacceleration()
	local current_velocity 		= entity.object:getvelocity()
	local now					= animals_get_current_time()
	local y_default_accel 		= animals_get_default_acceleration(pos,
									entity.animals_mpattern.movement_environment,
									entity.animals_mpattern.movement_canfly)


	-- use return codes of position check for initialization
	-- as initialization has already been done there
	local accel_changed = retval.accel_changed
	local force_change_of_speed = retval.force_accel_change
	local accel_to_set = retval.accel_to_set

	if accel_changed then
		minetest.log(LOGLEVEL_WARNING,"ANIMALS WARN:position got fixed ".. entity.data.name)
	end
	
	
	--predict next block animal will be
	local pos_predicted = movement_gen.predict_next_block(pos,current_velocity,current_acceleration)
	local pos_predicted_state = movement_gen.pos_is_ok(pos_predicted,entity)


	if accel_changed == false then
		accel_changed , accel_to_set = movement_gen.handle_level_change_end(entity,
												y_default_accel,
												current_velocity,
												current_acceleration,
												now,
												accel_to_set)
	end

	--next block animal is to be isn't a place where it can be so we need to change something
	if pos_predicted_state ~= "ok" then

		-- animal would walk onto water
		if accel_changed == false and pos_predicted_state == "above_water" then
			dbg_animals.movement_lvl1("animal is going to walk on water")
			local new_pos = movement_gen.get_suitable_pos_same_level(pos,1,entity)

			if new_pos ~= nil then
				
				accel_to_set = movement_gen.get_accel_to(new_pos,entity)
				accel_changed = true	
			else
				minetest.log(LOGLEVEL_ERROR,"ANIMALS: BUG!!! didn't find a way to stop animal at"..printpos(pos)..
								"from running into water")
				--TODO delete animal??
			end			
		end

		if accel_changed == false and pos_predicted_state == "collision_jumpable" then
			if movement_gen.pos_is_ok({x=pos_predicted.x,y=pos_predicted.y+1,z=pos_predicted.z},entity) == "ok" then
				if math.random() < entity.animals_mpattern.jump_up then
					local node_at_predicted_pos = minetest.env:get_node(pos_predicted)
					dbg_animals.movement_lvl2("ANIMALS: velocity is:" .. printpos(entity.object:getvelocity()) .. " position is: "..printpos(pos) ) 
					dbg_animals.movement_lvl2("ANIMALS: estimated position was: "..printpos(pos_predicted))
					dbg_animals.movement_lvl2("ANIMALS: predicted node state is: " .. movement_gen.pos_is_ok(pos_predicted,entity))
					--if node_at_predicted_pos ~= nil then
						--dbg_animals.movement_lvl1("ANIMALS: jumping onto: " .. node_at_predicted_pos.name)
					--end
					accel_to_set = { x=0,y=nil,z=0 }
					accel_changed = true
					
					local jumppos = {x=pos_predicted.x,y=centerpos.y+1,z=pos_predicted.z}
					
					dbg_animals.movement_lvl2("ANIMALS: animal " ..entity.data.name .. " is jumping, moving to:" .. printpos(jumppos))
					dbg_animals.movement_lvl2("ANIMALS: target pos node state is: " .. movement_gen.pos_is_ok(jumppos,entity))
					
					entity.object:moveto(jumppos)
				end
			end
		end

		--redirect animal to block thats not above its current level
		--or jump if possible
		if accel_changed == false and pos_predicted_state == "collision" then			
			dbg_animals.movement_lvl1("ANIMALS: animal is about to collide")

			local new_pos = movement_gen.get_suitable_pos_same_level(pos,1,entity)
		
			if new_pos ~= nil then
				
				accel_to_set = movement_gen.get_accel_to(new_pos,entity)
				accel_changed = true	
			end
		end

		--generic try to solve situation
		if accel_changed == false then
			accel_to_set = movement_gen.changeaccel(pos,prediction_time,
										entity)
			if accel_to_set ~= nil then
				accel_changed = true	
			end
		end
	end

	--avoid animals racing away
	local xzspeed = math.sqrt(math.exp(current_velocity.x,2)+math.exp(current_velocity.z,2))

	if xzspeed > entity.data.movement.max_speed then
		local newaccel = {x=0,y=current_acceleration.y,z=0}
		
		--calculate sign of acceleration
		if current_velocity.x > 0 and current_acceleration.x >= 0 then
				newaccel.x = entity.data.movement.min_accel * -1
		else
				newaccel.x = entity.data.movement.min_accel
		end
		
		if current_velocity.z > 0 and current_acceleration.z >= 0 then
				newaccel.z = entity.data.movement.min_accel * -1
		else
				newaccel.z = entity.data.movement.min_accel
		end
		
		--calculate relative partition of acceleration based on velocity
		if current_velocity.x > 0 and
			current_velocity.z > 0 then			
			newaccel.x = newaccel.x * current_velocity.x / (current_velocity.x+current_velocity.z)
			newaccel.z = newaccel.z * current_velocity.z / (current_velocity.x+current_velocity.z)			
		end
				
		--set acceleration of 25% of current speed
		accel_to_set = newaccel
		accel_changed = true
	end
	
	--this ain't perfect to avoid flying animals standing in air 
	--but it's a quick solution to fix most of the problems
	if entity.data.movement.min_speed ~= nil and
		xzspeed < entity.data.movement.min_speed then
		force_change_of_speed = true
	end

	--do randomized changes if not fighting
	if entity.dynamic_data.combat.target == "" then

		--apply random acceleration changes
		if accel_changed == false and
			(math.random() < entity.animals_mpattern.random_acceleration_change or 
			force_change_of_speed) then		
	
			accel_to_set = movement_gen.changeaccel(pos,prediction_time,
								entity)
			if accel_to_set ~= nil then
				accel_changed = true	
			end
			dbg_animals.movement_lvl1("ANIMALS: randomly changing speed from "..printpos(speed).." to "..printpos(speed_to_set))
		end
		
		
		--if animal is flying it's height may need to be changed due to changing ground level
		--calculate additional height change chances based on that value
		local below_min_chance = 0
		local above_max_chance = 0
		local override_timestamp = false
		if entity.animals_mpattern.movement_canfly then
			local min_y, max_y = animals_min_max_y_pos(
							entity.animals_mpattern.movement_environment,
							entity.animals_mpattern.movement_min_level,
							entity.animals_mpattern.movement_max_level,
							pos)
							
			if (pos.y < min_y) then
				if (min_y - pos.y) > 10 then
					below_min_chance = 1
					override_timestamp = true
				else
					below_min_change = (min_y - pos.y)/10
				end
			end
			
			if (pos.y > max_y) then
				if (pos.y - max_y) > 10 then
					above_max_chance = 1
					override_timestamp = true
				else
					above_max_change = (pos.y - max_y)/10
				end
			end	
		end
				

		--apply random jumps
		if accel_changed == false and
			(math.random() < entity.animals_mpattern.random_jump_chance or
			 math.random() < below_min_chance or
			 math.random() > above_max_chance ) then

			dbg_animals.movement_lvl1("ANIMALS: Random jump chance for animal " .. entity.data.name .. " "..
							entity.dynamic_data.movement.ts_random_jump .. " "..
							entity.animals_mpattern.random_jump_delay .. " " .. now
				)


			--flying/swiming animals do level change on random jump chance
			if entity.animals_mpattern.movement_canfly then
				
				local level_change_time = math.sqrt(2/entity.data.movement.min_accel)
			
				if now > (entity.dynamic_data.movement.ts_random_jump + 
					entity.animals_mpattern.random_jump_delay +
					level_change_time) or
					override_timestamp then

					if entity.dynamic_data.movement.changing_levels  == false or
						override_timestamp then
					
						accel_to_set = entity.object:getacceleration()
		
						local ydirection = 1
		
						if math.random() <= 0.5 then
							ydirection = -1
						end
		
						local targetpos = {x=pos.x,y=pos.y+ydirection,z=pos.z}
		
						local target_state = movement_gen.pos_is_ok(targetpos,entity);
		
						dbg_animals.movement_lvl2("ANIMALS: " .. entity.data.name .." flying change y accel dir="..ydirection.." ypos="..pos.y..
								" min="..entity.animals_mpattern.movement_min_level..
								" max="..entity.animals_mpattern.movement_max_level..
								" below="..target_state)
		
						if (target_state == "ok") then
						
							local min_y, max_y = animals_min_max_y_pos(
											entity.animals_mpattern.movement_environment,
											entity.animals_mpattern.movement_min_level,
											entity.animals_mpattern.movement_max_level,
											pos)
						
							dbg_animals.movement_lvl2("ANIMALS: check level borders current=".. pos.y .. 
														" min=" .. min_y ..
														" max=" .. max_y)
							if (pos.y > min_y and
								ydirection == -1) or
								(pos.y < max_y and
								ydirection == 1) 
								then
									accel_to_set.y = ydirection * entity.data.movement.min_accel
									accel_changed = true
									entity.dynamic_data.movement.ts_random_jump = now
									entity.dynamic_data.movement.changing_levels = true
									dbg_animals.movement_lvl2("ANIMALS: " .. entity.data.name .. " " .. now .. " " .. 
														" changing level within borders: " .. accel_to_set.y)
								end
						end
					else
						print("ANIMALS BUG!!! animal shouldn't be changing levels after that time!")
					end
				end
			--ground animals
			else
				if y_default_accel == 0 then
					minetest.log(LOGLEVEL_ERROR,"ANIMALS BUG!!! ground animal can't have zero default acceleration!!!")
				end
				
				local random_jump_time = 2 * (entity.animals_mpattern.random_jump_initial_speed /
												math.abs(y_default_accel))
			
				local next_jump = entity.dynamic_data.movement.ts_random_jump + 
									entity.animals_mpattern.random_jump_delay +
									random_jump_time
									
				dbg_animals.movement_lvl2("ANIMALS: now=" .. now .. " time of next jump=" .. next_jump ..
							" | " .. random_jump_time .. " " .. entity.animals_mpattern.random_jump_delay .. " " ..
							entity.dynamic_data.movement.ts_random_jump .. " " .. y_default_accel .. " " .. entity.animals_mpattern.random_jump_initial_speed)
			
				if now > next_jump then
				
					local ground_distance = animals_ground_distance(pos)
				
					
					--we shouldn't be moving in y direction at that time so probably we are just dropping
					--we can only jump while on solid ground!
					if ground_distance <= 1 then
					
						local current_speed = entity.object:getvelocity();

						dbg_animals.movement_lvl2("ANIMALS: " .. entity.data.name .." doing random jump ".. 
											"speed: " .. entity.animals_mpattern.random_jump_initial_speed ..
											": ",entity)
		
						current_speed.y = entity.animals_mpattern.random_jump_initial_speed
						
						entity.object:setvelocity(current_speed)
	
						entity.dynamic_data.movement.ts_random_jump = now
					else
						dbg_animals.movement_lvl2("ANIMALS: " .. entity.data.name .. " Random jump but ground distance was:" .. ground_distance .. 
										" " ..current_acceleration.y ..
										" " ..y_default_accel)
										
						--this is a workaround for a bug 
						--setting y acceleration to 0 for ground born animals shouldn't be possible
						if current_acceleration.y == 0 then
							accel_to_set = current_acceleration
							
							accel_to_set.y = y_default_accel
							
							accel_changed = true						
						end
					end
				end				
			end
		end
	end

	--really set new acceleration to entity
	if accel_changed then
		if accel_to_set == nil then
			accel_to_set = {x=0,
							y=y_default_accel,
							z=0}
			entity.object:setvelocity({x=0,y=0,z=0})
			minetest.log(LOGLEVEL_ERROR,"ANIMALS BUG!!!! set accel requested but no accel given!")
		end

		if accel_to_set.y == nil then
			accel_to_set.y = current_acceleration.y
		end

		-- make sure non flying animals have default acceleration
		if entity.animals_mpattern.movement_canfly == false then
			accel_to_set.y = y_default_accel
			
			if y_default_accel == 0 then
				print("ANIMALS BUG!!! " .. entity.data.name .. " animal that can't fly has acceleration 0!")
			end
		end


		--if math.abs (accel_to_set.y - y_default_accel) > 0.001 then
			--dbg_animals.movement_lvl1("Setting y-acceleration to: ".. accel_to_set.y .. " default:".. y_default_accel .. " name " .. entity.animals_name .. " at: " .. printpos(entity.object:getpos()))
		--end

		dbg_animals.movement_lvl1("ANIMALS: setting acceleration to " ..printpos(accel_to_set) )
		entity.dynamic_data.movement.acceleration = accel_to_set
		entity.object:setacceleration(accel_to_set)
	end
end

-------------------------------------------------------------------------------
-- name: callback(entity)
--
-- action: generate a movement pattern with mixed stop and go phases
--
-- param1: entity to generate movement
-- retval: -
-------------------------------------------------------------------------------
function movement_gen.callback(entity)
	if entity.dynamic_data.movement.started then

		local now     = animals_get_current_time()
		local delta_t = now - entity.dynamic_data.movement.ts_state_changed
		
		--case we are not moving
		if entity.dynamic_data.movement.moving == false then			
			
			local start_chance = entity.animals_mpattern.start_movement + 
									( 	delta_t * 
										entity.animals_mpattern.start_stop_delta_time_factor * 
										entity.animals_mpattern.start_movement
										)

			if math.random() < start_chance then
				dbg_animals.movement_lvl2("ANIMALS: animal starts to move, chance was:" .. start_chance .. 
											" deltatime was: " .. delta_t .. 
											" entity:", entity)
				
				--give movement gen chance to start movement
				movement_gen.generator(entity)
				entity.dynamic_data.movement.moving = true
				entity.dynamic_data.movement.ts_state_changed = now
			end
		else

			local stop_chance = entity.animals_mpattern.stop_movement + 
									( 	delta_t * 
										entity.animals_mpattern.start_stop_delta_time_factor * 
										entity.animals_mpattern.stop_movement)

			--random chance of stopping
			if math.random() < stop_chance then
				dbg_animals.movement_lvl2("ANIMALS: animal stops to move, chance was:" .. stop_chance .. 
											" deltatime was: " .. delta_t .. 
											" entity:", entity)
											
				entity.dynamic_data.movement.acceleration = {x=0,
												y=animals_get_default_acceleration(entity.object:getpos(),
														entity.animals_mpattern.movement_environment,
														entity.animals_mpattern.movement_canfly),
												z=0}
				entity.object:setacceleration(entity.dynamic_data.movement.acceleration)
				entity.object:setvelocity({x=0,y=0,z=0})
				entity.dynamic_data.movement.moving = false
				entity.dynamic_data.movement.ts_state_changed = now
			else
				movement_gen.generator(entity)
			end
		end


	else
		entity.object:setacceleration(entity.dynamic_data.movement.acceleration)
		entity.dynamic_data.movement.started = true
	end

end

-------------------------------------------------------------------------------
-- name: init_dynamic_data(entity,now)
--
-- action: initialize dynamic data required by movement generator
--
-- param1: entity to initialize
-- param2: current time
-- retval: -
-------------------------------------------------------------------------------
function movement_gen.init_dynamic_data(entity,now)

	local accel_to_set = {x=0,y=9.81,z=0}
	local pos = entity.object:getpos()

		--initialize acceleration values
	accel_to_set.y = animals_get_default_acceleration(pos,
					entity.animals_mpattern.movement_environment,
					entity.animals_mpattern.movement_canfly)

	local data = {
			started				= false,
			acceleration		= accel_to_set,
			moving				= false,
			changing_levels     = false,
			ts_state_changed	= now,
			ts_random_jump		= now,
			ts_orientation_upd  = now,
			}
	
	entity.dynamic_data.movement = data
end

-------------------------------------------------------------------------------
-- name: update_orientation(entity)
--
-- action: calculate direction to face
--
-- param1: entity to calculate direction
-- retval: -
-------------------------------------------------------------------------------

function update_orientation(entity)

	local now = animals_get_current_time()
	local new_orientation = 0

	if entity.dynamic_data.movement.ts_orientation_upd + 1 < now and
		entity.dynamic_data.movement.moving then

		entity.dynamic_data.movement.ts_orientation_upd = now

		current_velocity = entity.object:getvelocity()

		local x_abs = math.abs(current_velocity.x)
		local z_abs = math.abs(current_velocity.z)

		if x_abs > z_abs then
			if current_velocity.x > 0 then
				entity.object:setyaw(0)
			else
				entity.object:setyaw(3.14)
			end
		else
			if current_velocity.z >0 then
				entity.object:setyaw(1.57)
			else
				entity.object:setyaw(4,71)
			end
		end
	end

end


